# -*- encoding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    Copyright (c) 2016 E-MIPS (http://www.e-mips.com.ar)
#    Copyright (c) 2016 Eynes (http://www.eynes.com.ar)
#    All Rights Reserved. See AUTHORS for details.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General
#    Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
###############################################################################

from openerp.osv import fields, orm, osv
from openerp.tools.translate import _

import logging

_logger = logging.getLogger(__name__)


class GenerateComponent(orm.TransientModel):
    _name = 'generate.component'
    _description = 'Generate Component'

    _columns = {
        'system_id': fields.many2one('service.system', 'System'),
        'new_system_name': fields.char('New System Name'),
        'partner_id': fields.many2one('res.partner', 'Customer',
                                      readonly=True),
        'sale_line_ids': fields.one2many('generate.component.sale.line',
                                         'generate_component_id',
                                         'Sale Lines'),
        'note': fields.text('Note', readonly=True),
    }

    _defaults = {
        'sale_line_ids': lambda s, cr, uid, c: s._sale_line_default(
            cr, uid, c)[0],
        'note': lambda s, cr, uid, c: s._sale_line_default(
            cr, uid, c)[1],
        'partner_id': lambda s, cr, uid, c: c.get('partner_id', False),
    }

    def _get_sale(self, cr, uid, context=None):
        if not context:
            return []
        sale_obj = self.pool['sale.order']
        sale_id = context.get('active_ids', False)
        sale = sale_obj.browse(cr, uid, sale_id, context)
        if not sale_id or len(sale_id) != 1:
            raise osv.except_osv(
                _('Error'), _('Expected only one sale in Wizard Call!'))
        if isinstance(sale, list):
            sale = sale[0]
        return sale

    def _sale_line_default(self, cr, uid, context=None):
        wiz_line_obj = self.pool['generate.component.sale.line']
        sale = self._get_sale(cr, uid, context)
        line_ids = []
        note = []
        for line in sale.order_line:
            if not line.product_id.can_be_component:
                note.append(
                    _('Product `%s` can not be a component by previous ' +
                      'configuration.') % (line.product_id.name_template))
                continue
            remain = self._compute_remaining(cr, uid, line, context=context)
            if remain < 1:
                note.append(
                    _('Line with product `%s` has already generated ' +
                      '%s components.') % (line.product_id.name_template,
                                           int(line.product_uom_qty - remain)))
                continue
            for i in range(int(remain)):
                vals = {
                    'sale_line_id': line.id,
                }
                line_ids.append(wiz_line_obj.create(cr, uid, vals))
        return line_ids, '\n'.join(note)

    def _compute_remaining(self, cr, uid, line, context=None):
        component_obj = self.pool['service.component']
        generated_components = component_obj.search_count(
                cr, uid, [('sale_order_line', '=', line.id)])
        return line.product_uom_qty - generated_components

    def _check_lines(self, cr, uid, wiz, context=None):
        if not wiz.sale_line_ids:
            raise osv.except_osv(
                _("Error!"),
                _("There are no lines to make components from\n" +
                  "Please close the wizard and generate a new one."))
        return True

    def _generate_components_for_system(self, cr, uid, sale_lines,
                                        system, context=None):
        system_obj = self.pool['service.system']
        move_obj = self.pool['stock.move']
        component_obj = self.pool['service.component']
        new_components = []
        system_id = system.id
        s_lines = map(lambda x: x.sale_line_id, sale_lines)
        s_lines = list(set(s_lines))
        for s_l in s_lines:
            move_search = [
                ('sale_line_id', '=', s_l.id),
                # ('prodlot_id', '=', False),
                # ('state', 'not in', ['cancel', 'done'])
            ]
            move_ids = move_obj.search(cr, uid, move_search, context=context)
            comp_already_gen_ids = component_obj.search(
                cr, uid, [('move_id', 'in', move_ids)], context=context)
            comp_already_gen = component_obj.browse(
                cr, uid, comp_already_gen_ids, context=context)
            move_with_comp_ids = map(lambda x: x.move_id.id, comp_already_gen)
            move_ids = list(set(move_ids) - set(move_with_comp_ids))
            moves = move_obj.browse(cr, uid, move_ids, context=context)
            comps = filter(lambda x: x.sale_line_id == s_l, sale_lines)
            # if sum(map(lambda x: x.product_qty, moves)) < len(comps):
            #     raise osv.except_osv(
            #         _('Error'),
            #         _('Could not link product `%s` to `%s` stock moves\n' +
            #           'Maybe those moves are already done or have prodlot set')
            #         % (comps[0].sale_line_id.product_id.name_template,
            #            len(comps))
            #     )
            moves = sorted(moves, key=lambda x: x.product_qty)
            moves_map = []
            for index, move in enumerate(moves):
                moves_map.append({
                    'id': move.id,
                    'qty': move.product_qty,
                    'initial_qty': move.product_qty,
                    'prodlot_id': move.prodlot_id.id if move.prodlot_id else False,
                })
            for index, comp in enumerate(comps):
                new_components_vals = {
                        'product_id': s_l.product_id.id,
                        'partner_id': s_l.order_id.partner_id.id,
                        'system_id': system_id,
                        'sale_order_line': s_l.id,
                }
                if moves_map:
                    move = moves_map[0]
                    comp_move_id = move['id']
                    prodlot_id = move['prodlot_id']
                    if move['qty'] == 1:
                        moves_map.pop(0)
                        if move['initial_qty'] != 1:
                            vals = {
                                'product_qty': 1,
                            }
                            move_obj.write(cr, uid, [move['id']], vals,
                                           context=context)
                    else:
                        default = {
                            'product_qty': 1,
                            'state': move_obj.browse(
                                cr, uid, move['id'], context=context).state
                        }
                        comp_move_id = move_obj.copy(cr, uid, move['id'],
                                                     default, context=context)
                        move['qty'] -= 1

                    new_components_vals.update({
                        'move_id': comp_move_id,
                        'prodlot_id': prodlot_id,
                        })
                
                new_components.append(new_components_vals)
                
            if moves_map:
                # Substract the splitted qty from the last packed move
                last_move = moves_map[0]
                vals = {
                    'product_qty': last_move['qty'],
                }
                move_obj.write(cr, uid, [last_move['id']],
                               vals, context=context)
        write_vals = {
            'component_ids': [(0, 0, x) for x in new_components],
        }
        system_obj.write(cr, uid, system_id,
                         write_vals, context=context)

    def _create_system(self, cr, uid, wiz, context=None):
        system_pool = self.pool['service.system']
        sale = self._get_sale(cr, uid, context)
        create_vals = {
            'name': wiz.new_system_name,
            'partner_id': wiz.partner_id.id,
            'sector_id': sale.technique_sector_id.id,
        }
        system_id = system_pool.create(
            cr, uid, create_vals, context=context)
        return system_pool.browse(cr, uid, system_id, context)

    def _launch_system_action(self, cr, uid, system, context=None):
        act_window = {
            'name': _('System'),
            'res_id': system.id,
            'res_model': 'service.system',
            'target': 'current',
            'type': 'ir.actions.act_window',
            'view_type': 'form',
            'view_mode': 'form',
        }
        return act_window

    def generate_components(self, cr, uid, ids, context=None):
        wiz = self.browse(cr, uid, ids, context)[0]
        self._check_lines(cr, uid, wiz, context)
        self._generate_components_for_system(cr, uid, wiz.sale_line_ids,
                                             wiz.system_id, context=context)
        return self._launch_system_action(cr, uid, wiz.system_id, context)

    def create_new_system(self, cr, uid, ids, context=None):
        wiz = self.browse(cr, uid, ids, context)[0]
        self._check_lines(cr, uid, wiz, context)
        sys_id = self._create_system(cr, uid, wiz, context=context)
        self._generate_components_for_system(cr, uid, wiz.sale_line_ids,
                                             sys_id, context=context)
        return self._launch_system_action(cr, uid, sys_id, context)


class GenerateComponentSaleLine(orm.TransientModel):
    _name = 'generate.component.sale.line'
    _description = 'Generate Component Sale Line'

    _columns = {
        'generate_component_id': fields.many2one('generate.component',
                                                 'Generate Component'),
        'sale_line_id': fields.many2one('sale.order.line', 'Sale Line'),
        'product_id': fields.related('sale_line_id', 'product_id',
                                     type='many2one',
                                     relation='product.product',
                                     string='Product'),
    }
